"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of jerrycan.

jerrycan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jerrycan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jerrycan.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import pathlib
from unittest.mock import patch
from jerrycan.test.TestFramework import _TestFramework
from jerrycan.Config import Config


class TestConfig(_TestFramework):
    """
    Class that tests the config class
    """

    def test_db_config(self):
        """
        Tests the database configuration
        :return: None
        """
        os.environ.pop("FLASK_TESTING")
        os.environ["DB_MODE"] = "sqlite"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.DB_URI, "sqlite:////tmp/jerrycan.db")
        os.environ["SQLITE_PATH"] = "/tmp/test.db"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.DB_URI, "sqlite:////tmp/test.db")

        os.environ["DB_MODE"] = "mysql"
        os.environ["MYSQL_USER"] = "abc"
        os.environ["MYSQL_PASSWORD"] = "def"
        os.environ["MYSQL_HOST"] = "ghi"
        os.environ["MYSQL_PORT"] = "1000"
        os.environ["MYSQL_DATABASE"] = "xyz"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.DB_URI, "mysql://abc:def@ghi:1000/xyz")
        os.environ["DB_MODE"] = "postgresql"
        os.environ["POSTGRES_USER"] = "ABC"
        os.environ["POSTGRES_PASSWORD"] = "DEF"
        os.environ["POSTGRES_HOST"] = "GHI"
        os.environ["POSTGRES_PORT"] = "2000"
        os.environ["POSTGRES_DATABASE"] = "XYZ"

        try:
            Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
            self.fail()
        except SystemExit:
            pass

        os.environ["POSTGRES_DB"] = "XYZ"

        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.DB_URI, "postgresql://ABC:DEF@GHI:2000/XYZ")

    def test_version(self):
        """
        Tests if the version is fetched correctly
        :return: None
        """
        version_file = os.path.join(
            pathlib.Path(__file__).parent.absolute(),
            "../../../../jerrycan/version"
        )
        with open(version_file, "r") as f:
            version = f.read()
        self.assertEqual(version, Config.VERSION)

    def test_environment_variables_definitions(self):
        """
        Tests the definition of the environent variables
        :return: None
        """
        if "DB_MODE" in os.environ:
            os.environ.pop("DB_MODE")
        variables = Config.environment_variables()
        first_required = variables["required"]
        first_optional = variables["optional"]
        self.assertTrue("FLASK_SECRET" in first_required)
        self.assertFalse("SQLITE_PATH" in first_optional)
        self.assertFalse("MYSQL_USER" in first_required)
        self.assertFalse("MYSQL_DATABASE" in first_required)
        self.assertFalse("MYSQL_DB" in first_required)
        self.assertFalse("POSTGRES_USER" in first_required)
        self.assertFalse("POSTGRES_DATABASE" in first_required)
        self.assertFalse("POSTGRES_DB" in first_required)
        self.assertTrue("LOGGING_PATH" in first_optional)

        os.environ["DB_MODE"] = "sqlite"
        variables = Config.environment_variables()
        required = variables["required"]
        optional = variables["optional"]
        self.assertEqual(first_required, required)
        self.assertNotEqual(first_optional, optional)
        self.assertTrue("SQLITE_PATH" in optional)
        self.assertFalse("MYSQL_USER" in required)
        self.assertFalse("MYSQL_DATABASE" in required)
        self.assertFalse("MYSQL_DB" in required)
        self.assertFalse("POSTGRES_USER" in required)
        self.assertFalse("POSTGRES_DATABASE" in required)
        self.assertFalse("POSTGRES_DB" in required)

        os.environ["DB_MODE"] = "mysql"
        variables = Config.environment_variables()
        required = variables["required"]
        optional = variables["optional"]
        self.assertNotEqual(first_required, required)
        self.assertEqual(first_optional, optional)
        self.assertFalse("SQLITE_PATH" in optional)
        self.assertTrue("MYSQL_USER" in required)
        self.assertTrue("MYSQL_DATABASE" in required)
        self.assertFalse("MYSQL_DB" in required)
        self.assertFalse("POSTGRES_USER" in required)
        self.assertFalse("POSTGRES_DATABASE" in required)
        self.assertFalse("POSTGRES_DB" in required)

        os.environ["DB_MODE"] = "postgresql"
        variables = Config.environment_variables()
        required = variables["required"]
        optional = variables["optional"]
        self.assertNotEqual(first_required, required)
        self.assertEqual(first_optional, optional)
        self.assertFalse("SQLITE_PATH" in optional)
        self.assertFalse("MYSQL_USER" in required)
        self.assertFalse("MYSQL_DATABASE" in required)
        self.assertFalse("MYSQL_DB" in required)
        self.assertTrue("POSTGRES_USER" in required)
        self.assertFalse("POSTGRES_DATABASE" in required)
        self.assertTrue("POSTGRES_DB" in required)

    def test_dumping_environment_variables(self):
        """
        Tests dumping environment variables
        :return: None
        """
        dumpfile = "/tmp/envdump"
        os.environ["FLASK_SECRET"] = "ABCXYZ"
        Config.dump_env_variables(dumpfile)

        def verify(output: str):
            with open(dumpfile, "r") as f:
                self.assertEqual(output, f.read())

        with patch("builtins.print", verify):
            Config.dump_env_variables()

    def test_building_base_url(self):
        """
        Tests building the base URL
        :return: None
        """
        os.environ["DOMAIN_NAME"] = "example.org"
        os.environ["HTTP_PORT"] = "1000"
        os.environ["BEHIND_PROXY"] = "0"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.base_url(), "http://example.org:1000")

        os.environ["BEHIND_PROXY"] = "1"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")
        self.assertEqual(Config.base_url(), "https://example.org")

    def test_initializing_telegram_bot_connection(self):
        """
        'Tests' initializing the telegram bot connection
        :return: None
        """
        class DummyBot:
            def __init__(self, settings):
                self.settings = settings

        os.environ["TELEGRAM_API_KEY"] = "ABC"
        Config.load_config(self.root_path, "jerrycan", "sentry_dsn")

        try:
            self.assertIsNone(Config.TELEGRAM_BOT_CONNECTION)
            self.fail()
        except AttributeError:
            pass

        with patch("jerrycan.Config.TelegramBotConnection", DummyBot):
            Config.initialize_telegram()
        self.assertIsNotNone(Config.TELEGRAM_BOT_CONNECTION)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of jerrycan.

jerrycan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jerrycan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jerrycan.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from enum import Enum
from typing import Dict, Any
from jerrycan.base import db
from jerrycan.db.ModelMixin import ModelMixin
from jerrycan.test.TestFramework import _TestFramework


class TestModelMixin(_TestFramework):
    """
    Class that tests the ModelMixin class
    """

    def test_enum_attributes(self):
        """
        Tests if enum attributes are handled correctly
        :return: None
        """
        class A(Enum):
            B = 1
            C = 2

        class Tester(ModelMixin, db.Model):
            enum = db.Column(db.Enum(A))

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

        tester = Tester(id=1, enum=A.B)
        self.assertEqual(repr(tester), "Tester(id=1, enum=A.B)")
        self.assertEqual(tester.__json__()["enum"], "B")

    def test_json_representation(self):
        """
        Tests the JSOn representation of a ModelMixin
        :return: None
        """
        class A(ModelMixin, db.Model):
            __tablename__ = "a"
            s = db.Column(db.String(255))

        class B(ModelMixin, db.Model):
            __tablename__ = "b"
            a_id = db.Column(db.Integer, db.ForeignKey("a.id"))
            a = db.relationship("A")
            user_id: int = db.Column(
                db.Integer,
                db.ForeignKey("users.id"),
                nullable=False
            )

        # noinspection PyArgumentList
        b1 = B(a=A(), a_id=1)
        # noinspection PyArgumentList
        b2 = B(a=None)
        self.assertNotEqual(b1.__json__(), b2.__json__())
        self.assertNotEqual(b1.__json__(True), b2.__json__(True))



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of jerrycan.

jerrycan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jerrycan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jerrycan.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
from bokkichat.entities.Address import Address
from bokkichat.entities.message.TextMessage import TextMessage
from jerrycan.test.TestFramework import _TestFramework


class TestTelegramChatId(_TestFramework):
    """
    Class that tests the TelegramChatId database model
    """

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)
        self.assertEqual(
            chat_id.__json__(False),
            {
                "id": chat_id.id,
                "user_id": user.id,
                "chat_id": chat_id.chat_id
            }
        )
        user_json = user.__json__(True)
        user_json.pop("telegram_chat_id")
        self.assertEqual(
            chat_id.__json__(True),
            {
                "id": chat_id.id,
                "user_id": user.id,
                "user": user_json,
                "chat_id": chat_id.chat_id
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)
        data = chat_id.__json__()
        data.pop("id")
        self.assertEqual(
            str(chat_id),
            "TelegramChatId:{} <{}>".format(chat_id.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        user = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)
        generated = {"value": chat_id}
        code = repr(chat_id)

        # noinspection PyUnresolvedReferences
        from jerrycan.db.TelegramChatId import TelegramChatId
        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], chat_id)
        self.assertFalse(generated["value"] is chat_id)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        user = self.generate_sample_user()[0]
        user2 = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)
        chat_id2 = self.generate_telegram_chat_id(user2)
        mapping = {
            chat_id: 100,
            chat_id2: 200
        }
        self.assertEqual(mapping[chat_id], 100)
        self.assertEqual(mapping[chat_id2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        user = self.generate_sample_user()[0]
        user2 = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)
        chat_id2 = self.generate_telegram_chat_id(user2)
        self.assertEqual(chat_id, chat_id)
        self.assertNotEqual(chat_id, chat_id2)
        self.assertNotEqual(chat_id, 100)

    def test_sending_message(self):
        """
        Tests sending a message
        :return: None
        """
        user = self.generate_sample_user()[0]
        chat_id = self.generate_telegram_chat_id(user)

        class DummyConnection:
            address = Address("ABC")
            parent: _TestFramework = self

            def send(self, message: TextMessage):
                self.parent.assertEqual(message.sender.address,
                                        self.address.address)
                self.parent.assertEqual(message.receiver.address,
                                        chat_id.chat_id)
                self.parent.assertEqual(message.body, "XYZ")

        class DummyConfig:
            TELEGRAM_BOT_CONNECTION = DummyConnection()

        with patch("jerrycan.db.TelegramChatId.Config", DummyConfig):
            chat_id.send_message("XYZ")



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of jerrycan.

jerrycan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jerrycan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jerrycan.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
# noinspection PyPackageRequirements
from telegram.error import Conflict
from bokkichat.entities.Address import Address
from bokkichat.entities.message.MediaMessage import MediaMessage, MediaType
from bokkichat.entities.message.TextMessage import TextMessage
from jerrycan.test.TestFramework import _TestFramework
from jerrycan.background.telegram import telegram_whoami


class TestTelegramBackground(_TestFramework):
    """
    Tests the telegram background functions
    """

    def test_whoami(self):
        """
        Tests the whoami functionality
        :return: None
        """
        class DummyConnection:
            address = Address("ABC")
            parent: _TestFramework = self
            create_errors = False
            counter = 0

            def send(self, message: TextMessage):

                if self.create_errors:
                    raise Conflict("Error")

                self.counter += 1

                self.parent.assertEqual(message.sender.address,
                                        self.address.address)
                self.parent.assertEqual(message.body, message.receiver.address)

            def loop(self, handler):
                handler(None, TextMessage(
                    Address("XYZ"),
                    self.address,
                    "/whoami"
                ))
                handler(None, TextMessage(
                    Address("XYZ"),
                    self.address,
                    "Hello"
                ))
                handler(None, MediaMessage(
                    Address("XYZ"),
                    self.address,
                    MediaType.AUDIO,
                    b""
                ))

        connection = DummyConnection()

        class DummyConfig:
            TELEGRAM_BOT_CONNECTION = connection

        with patch("jerrycan.background.telegram.Config", DummyConfig):
            self.assertEqual(connection.counter, 0)
            telegram_whoami()
            self.assertEqual(connection.counter, 1)
            DummyConfig.TELEGRAM_BOT_CONNECTION.create_errors = True
            telegram_whoami()
            self.assertEqual(connection.counter, 1)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of jerrycan.

jerrycan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jerrycan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jerrycan.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
from jerrycan.test.TestFramework import _TestFramework
from jerrycan.exceptions import ApiException


class TestErrorHandling(_TestFramework):
    """
    Class that tests the flask error handling
    """

    def test_exception(self):
        """
        Tests if unexpected exceptions are caught correctly
        :return: None
        """

        def render_template(*_, **__):
            print(1/0)
        with patch("jerrycan.routes.static.render_template",
                   render_template):
            resp = self.client.get("/", follow_redirects=True)
            self.assertTrue(b"The server encountered an internal" in resp.data)

    def test_404(self):
        """
        Tests if a 404 error is handled correctly
        :return: None
        """
        resp = self.client.get("/baba", follow_redirects=True)
        print(resp.data)
        self.assertTrue(b"404 Not Found" in resp.data)

    def test_api_exception(self):
        """
        Tests if ApiExceptions in the API are handled correctly
        :return: None
        """
        class RequestDummy:
            @staticmethod
            def get_json():
                raise ApiException("Test", 500)

        with patch("jerrycan.routes.api.user_management.request",
                   RequestDummy):
            resp = self.client.post(
                "/api/v1/key", follow_redirects=True, json={}
            )
            self.assertEqual(resp.status_code, 500)
